/**
 * Title: CombineGPX
 * Description: Combines <gpxtpx:hr> values into another GPX
 * @author: Sven Killig &lt;sven@killig.de&gt;
 * Requires XOM from http://www.xom.nu
 * HB/5 m
 * 
 */

package de.killig;

import java.awt.Component;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.StringSelection;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.InputStream;
import java.io.PrintWriter;
import java.text.DateFormat;
import java.time.Instant;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.prefs.Preferences;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileFilter;

import com.garmin.fit.BatteryStatus;
import com.garmin.fit.Decode;
import com.garmin.fit.DeveloperField;
import com.garmin.fit.DeveloperFieldDescription;
import com.garmin.fit.DeveloperFieldDescriptionListener;
import com.garmin.fit.DeviceInfoMesg;
import com.garmin.fit.DeviceInfoMesgListener;
import com.garmin.fit.Factory;
import com.garmin.fit.Field;
import com.garmin.fit.FieldBase;
import com.garmin.fit.FileIdMesg;
import com.garmin.fit.FileIdMesgListener;
import com.garmin.fit.Fit;
import com.garmin.fit.FitRuntimeException;
import com.garmin.fit.Gender;
import com.garmin.fit.Mesg;
import com.garmin.fit.MesgBroadcaster;
import com.garmin.fit.MonitoringMesg;
import com.garmin.fit.MonitoringMesgListener;
import com.garmin.fit.RecordMesg;
import com.garmin.fit.RecordMesgListener;
import com.garmin.fit.UserProfileMesg;
import com.garmin.fit.UserProfileMesgListener;

import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Elements;
import nu.xom.Node;


//TODO replace with TrkPtInt
class TrkPt implements Comparable<TrkPt> {
	ZonedDateTime zdt;
	Element hr;
	Element temp;

	public TrkPt(ZonedDateTime zdt, Element hr, Element temp) {
		this.zdt=zdt;
		this.hr=hr;
		this.temp=temp;
	}
	
	public int compareTo(TrkPt o) {
		return zdt.compareTo(o.zdt);
	}
	
	public String toString() {
		return zdt+"\t"+hr.getValue()+(temp!=null?"\t"+temp.getValue():"");
	}
}

class TrkPtInt implements Comparable<TrkPtInt> {
	Instant zdt;
	int value;

	public TrkPtInt(Instant zdt, int value) {
		this.zdt=zdt;
		this.value=value;
	}
	
	public int compareTo(TrkPtInt o) {
		return zdt.compareTo(o.zdt);
	}
	
	public String toString() {
		return zdt+"\t"+value;
	}
}


class JFileChooserPref extends JFileChooser {
  Preferences prefs;

  public JFileChooserPref() {
		super();
		Class callerClass=StackWalker.getInstance(StackWalker.Option.RETAIN_CLASS_REFERENCE).getCallerClass();
//	  System.out.println("callerClass="+callerClass);
	  prefs=Preferences.userNodeForPackage(callerClass);
	  Class clazz=this.getClass();
//	  System.out.println("clazz="+clazz);
	  String prefPath=prefs.get(clazz.getSimpleName(), ".");
//	  System.out.println("prefPath="+prefPath);
		setCurrentDirectory(new File(prefPath));
	}
  
  public int showOpenDialog(Component parent) {
  	int returnVal=super.showOpenDialog(parent);
//	  System.out.println("returnVal="+returnVal);
    if (returnVal == JFileChooser.APPROVE_OPTION) {
      prefs.put(this.getClass().getSimpleName(), getCurrentDirectory().getAbsolutePath());
    }
    return returnVal;
  }
}


public class CombineGPX {
	static int hr;
	static int hrMin=Integer.MAX_VALUE, hrMax=Integer.MIN_VALUE;
	static int middle;

	static void appendHRs(Document doc, List<TrkPt> list) {
    Element gpx = doc.getRootElement();
    Element trk = gpx.getFirstChildElement("trk", "http://www.topografix.com/GPX/1/1");
    Element trkseg = trk.getFirstChildElement("trkseg", "http://www.topografix.com/GPX/1/1");
    Elements trkpts = trkseg.getChildElements("trkpt", "http://www.topografix.com/GPX/1/1");
    
    for (int i = 0; i < trkpts.size(); i++) {
    	Element trkpt=trkpts.get(i);
      String time = trkpt.getFirstChildElement("time", "http://www.topografix.com/GPX/1/1").getValue();
      ZonedDateTime zdatetime = ZonedDateTime.parse(time, DateTimeFormatter.ISO_DATE_TIME);
      Element hrElem=trkpt.getFirstChildElement("extensions", "http://www.topografix.com/GPX/1/1").getFirstChildElement("TrackPointExtension", "http://www.garmin.com/xmlschemas/TrackPointExtension/v1").getFirstChildElement("hr", "http://www.garmin.com/xmlschemas/TrackPointExtension/v1");
      Element tempElem=trkpt.getFirstChildElement("extensions", "http://www.topografix.com/GPX/1/1").getFirstChildElement("TrackPointExtension", "http://www.garmin.com/xmlschemas/TrackPointExtension/v1").getFirstChildElement("atemp", "http://www.garmin.com/xmlschemas/TrackPointExtension/v1");
      if(hrElem!=null) {
      	hr = Integer.parseInt(hrElem.getValue());
	      hrMin=Integer.min(hrMin, hr); hrMax=Integer.max(hrMax, hr);
	      TrkPt tp=new TrkPt(zdatetime, hrElem, tempElem);
	      list.add(tp);
	      System.out.println(tp);
      }
    }
    System.out.println(list.size());
    if(middle==0) middle=list.size()-1;
	}
	
	
	static List<TrkPtInt> listTemp;
	public static void decodeTempCSV(File arg) {
		try {
	  	BufferedReader buf=new BufferedReader(new FileReader(arg));
			String s;
			while((s=buf.readLine()) != null) {
				String[] as=s.split(",");
        TrkPtInt tpi=new TrkPtInt(Instant.parse(as[0]), Math.round(Float.parseFloat(as[1])));
        listTemp.add(tpi);
        System.out.println(tpi);
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	public static void decodeTempFIT(File arg) {
    Decode decode = new Decode();
    //decode.skipHeader();        // Use on streams with no header and footer (stream contains FIT defn and data messages only)
    //decode.incompleteStream();  // This suppresses exceptions with unexpected eof (also incorrect crc)
    MesgBroadcaster mesgBroadcaster = new MesgBroadcaster(decode);
    Listener listener = new Listener();
    FileInputStream in;

    System.out.printf("FIT Decoder - Protocol %d.%d Profile %.2f %s\n", Fit.PROTOCOL_VERSION_MAJOR, Fit.PROTOCOL_VERSION_MINOR, Fit.PROFILE_VERSION / 100.0, Fit.PROFILE_TYPE);

    try {
        in = new FileInputStream(arg);
    } catch (java.io.IOException e) {
        throw new RuntimeException("Error opening file " + arg + " [1]");
    }

    try {
        if (!decode.checkFileIntegrity((InputStream)in)) {
            throw new RuntimeException("FIT file integrity failed.");
        }
    } catch (RuntimeException e) {
        System.err.print("Exception Checking File Integrity: ");
        System.err.println(e.getMessage());
        System.err.println("Trying to continue...");
    } finally {
        try {
            in.close();
        } catch (java.io.IOException e) {
            throw new RuntimeException(e);
        }
    }

    try {
        in = new FileInputStream(arg);
    } catch (java.io.IOException e) {
        throw new RuntimeException("Error opening file " + arg + " [2]");
    }

    mesgBroadcaster.addListener((FileIdMesgListener)listener);
    mesgBroadcaster.addListener((UserProfileMesgListener)listener);
    mesgBroadcaster.addListener((DeviceInfoMesgListener)listener);
    mesgBroadcaster.addListener((MonitoringMesgListener)listener);
    mesgBroadcaster.addListener((RecordMesgListener)listener);

    decode.addListener((DeveloperFieldDescriptionListener)listener);

    try {
        decode.read(in, mesgBroadcaster, mesgBroadcaster);
    } catch (FitRuntimeException e) {
        // If a file with 0 data size in it's header  has been encountered,
        // attempt to keep processing the file
        if (decode.getInvalidFileDataSize()) {
            decode.nextFile();
            decode.read(in, mesgBroadcaster, mesgBroadcaster);
        } else {
            System.err.print("Exception decoding file: ");
            System.err.println(e.getMessage());

            try {
                in.close();
            } catch (java.io.IOException f) {
                throw new RuntimeException(f);
            }

            return;
        }
    }

    try {
        in.close();
    } catch (java.io.IOException e) {
        throw new RuntimeException(e);
    }

    System.out.println("Decoded FIT file " + arg);
}

  private static class Listener implements FileIdMesgListener, UserProfileMesgListener, DeviceInfoMesgListener, MonitoringMesgListener, RecordMesgListener, DeveloperFieldDescriptionListener {

    @Override
    public void onMesg(FileIdMesg mesg) {
        System.out.println("File ID:");

        if (mesg.getType() != null) {
            System.out.print("   Type: ");
            System.out.println(mesg.getType().getValue());
        }

        if (mesg.getManufacturer() != null ) {
            System.out.print("   Manufacturer: ");
            System.out.println(mesg.getManufacturer());
        }

        if (mesg.getProduct() != null) {
            System.out.print("   Product: ");
            System.out.println(mesg.getProduct());
        }

        if (mesg.getSerialNumber() != null) {
            System.out.print("   Serial Number: ");
            System.out.println(mesg.getSerialNumber());
        }

        if (mesg.getNumber() != null) {
            System.out.print("   Number: ");
            System.out.println(mesg.getNumber());
        }
    }

    @Override
    public void onMesg(UserProfileMesg mesg) {
        System.out.println("User profile:");

        if (mesg.getFriendlyName() != null) {
            System.out.print("   Friendly Name: ");
            System.out.println(mesg.getFriendlyName());
        }

        if (mesg.getGender() != null) {
            if (mesg.getGender() == Gender.MALE) {
                System.out.println("   Gender: Male");
            } else if (mesg.getGender() == Gender.FEMALE) {
                System.out.println("   Gender: Female");
            }
        }

        if (mesg.getAge() != null) {
            System.out.print("   Age [years]: ");
            System.out.println(mesg.getAge());
        }

        if (mesg.getWeight() != null) {
            System.out.print("   Weight [kg]: ");
            System.out.println(mesg.getWeight());
        }
    }

    @Override
    public void onMesg(DeviceInfoMesg mesg) {
        System.out.println("Device info:");

        if (mesg.getTimestamp() != null) {
            System.out.print("   Timestamp: ");
            System.out.println(mesg.getTimestamp());
        }

        if (mesg.getBatteryStatus() != null) {
            System.out.print("   Battery status: ");

            switch (mesg.getBatteryStatus()) {
            case BatteryStatus.CRITICAL:
                System.out.println("Critical");
                break;
            case BatteryStatus.GOOD:
                System.out.println("Good");
                break;
            case BatteryStatus.LOW:
                System.out.println("Low");
                break;
            case BatteryStatus.NEW:
                System.out.println("New");
                break;
            case BatteryStatus.OK:
                System.out.println("OK");
                break;
            default:
                System.out.println("Invalid");
                break;
            }
        }
    }

    @Override
    public void onMesg(MonitoringMesg mesg) {
        System.out.println("Monitoring:");

        if (mesg.getTimestamp() != null) {
            System.out.print("   Timestamp: ");
            System.out.println( mesg.getTimestamp());
        }

        if (mesg.getActivityType() != null) {
            System.out.print("   Activity Type: ");
            System.out.println(mesg.getActivityType());
        }

        // Depending on the ActivityType, there may be Steps, Strokes, or Cycles present in the file
        if (mesg.getSteps() != null) {
            System.out.print("   Steps: ");
            System.out.println( mesg.getSteps());
        } else if (mesg.getStrokes() != null) {
            System.out.print("   Strokes: ");
            System.out.println(mesg.getStrokes());
        } else if (mesg.getCycles() != null) {
            System.out.print("   Cycles: ");
            System.out.println(mesg.getCycles());
        }

        printDeveloperData(mesg);
    }

    @Override
    public void onMesg(RecordMesg mesg) {
        //System.out.println("Record:");

        printValues(mesg, RecordMesg.HeartRateFieldNum);
        printValues(mesg, RecordMesg.CadenceFieldNum);
        printValues(mesg, RecordMesg.DistanceFieldNum);
        printValues(mesg, RecordMesg.SpeedFieldNum);
        //System.out.println(mesg.getTimestamp());
        //System.out.println(mesg.getTemperature()+"�C");
        Calendar calendar = Calendar.getInstance();
        // 2018-11-16T16:58:17+01:00[Europe/Berlin] instead of 2018-11-16T16:58:17Z
        //GregorianCalendar calendar = new GregorianCalendar();
        calendar.setTime(mesg.getTimestamp().getDate());
        Instant i=calendar.toInstant();
        Byte b=mesg.getTemperature();
        if(b!=null) {
	        TrkPtInt tpi=new TrkPtInt(i, b);
	        listTemp.add(tpi);
	        System.out.println(tpi);
        }

        printDeveloperData(mesg);
    }

    private void printDeveloperData(Mesg mesg) {
        for (DeveloperField field : mesg.getDeveloperFields()) {
            if (field.getNumValues() < 1) {
                continue;
            }

            if (field.isDefined()) {
                System.out.print("   " + field.getName());

                if (field.getUnits() != null) {
                    System.out.print(" [" + field.getUnits() + "]");
                }

                System.out.print(": ");
            } else {
                System.out.print("   Undefined Field: ");
            }

            System.out.print(field.getValue( 0 ));
            for (int i = 1; i < field.getNumValues(); i++) {
                System.out.print("," + field.getValue(i));
            }

            System.out.println();
        }
    }

    @Override
    public void onDescription(DeveloperFieldDescription desc) {
        System.out.println("New Developer Field Description");
        System.out.println("   App Id: " + desc.getApplicationId());
        System.out.println("   App Version: " + desc.getApplicationVersion());
        System.out.println("   Field Num: " + desc.getFieldDefinitionNumber());
    }

    private void printValues(Mesg mesg, int fieldNum) {
        Iterable<FieldBase> fields = mesg.getOverrideField((short) fieldNum);
        Field profileField = Factory.createField(mesg.getNum(), fieldNum);
        boolean namePrinted = false;

        if (profileField == null) {
            return;
        }

        for (FieldBase field : fields) {
            if (!namePrinted) {
                System.out.println("   " + profileField.getName() + ":");
                namePrinted = true;
            }

            if (field instanceof Field) {
                System.out.println("      native: " + field.getValue());
            } else {
                System.out.println("      override: " + field.getValue());
            }
        }
    }
}

  
  static <T> int binarySearch(List<? extends Comparable<? super T>> list, T key) {
    int res=Collections.binarySearch(list, key);
    if(res<0) {
    	if(res==-1) {
    		// nop
    	} else if(res>(-list.size()+1)) {
    		// ">" because negative
    		res=Math.abs(res)-1-1;
    	} else {
    		// nop
    	}
    } else if(res>0) {
//        	res--;
    }
    return res;
  }
  
  public static void main(String[] args) throws Exception {
		List<File> argFiles=new ArrayList<File>();
		boolean tempe=false;
		
				if (args.length < 2) {
          System.out.println("Usage: java -jar "+CombineGPX.class.getSimpleName()+".jar [cobi.gpx hr.gpx [hr2.gpx ...] [tempe.fit|csv]]");
          final JFileChooser fc = new JFileChooserPref();
          fc.setFileFilter(new FileFilter() {
						@Override
						public boolean accept(File f) {
							return f.getName().endsWith(".gpx") || f.isDirectory();
						}
						@Override
						public String getDescription() {
							return "*.gpx";
						}
          });
          fc.setDialogTitle("COBI.Bike");
          int returnVal = fc.showOpenDialog(null);
          if (returnVal == JFileChooser.APPROVE_OPTION) {
          	argFiles.add(fc.getSelectedFile());
          } else {
            return;
          }
          fc.setDialogTitle("Heart Rate(s)");
          fc.setMultiSelectionEnabled(true);
          returnVal = fc.showOpenDialog(null);
          if (returnVal == JFileChooser.APPROVE_OPTION) {
          	argFiles.addAll(List.of(fc.getSelectedFiles()));
          } else {
            return;
          }

          fc.setFileFilter(new FileFilter() {
						@Override
						public boolean accept(File f) {
							return f.getName().endsWith(".fit") || f.getName().endsWith(".csv") || f.isDirectory();
						}
						@Override
						public String getDescription() {
							return "*.fit;*.csv";
						}
          });
          fc.setDialogTitle("tempe");
          fc.setMultiSelectionEnabled(false);
          returnVal = fc.showOpenDialog(null);
          if (returnVal == JFileChooser.APPROVE_OPTION) {
          	argFiles.add(fc.getSelectedFile());
          	tempe=true;
          }
        } else {
        	argFiles.add(new File(args[0]));
          for(int i=1; i<args.length; i++) {
          	argFiles.add(new File(args[i]));
          }
        }

        Builder builder = new Builder();

        List<TrkPt> list=new ArrayList<TrkPt>();
        for(int i=1; i<argFiles.size()-(tempe?1:0); i++) {
        	if(!(argFiles.get(i).getName().endsWith(".fit") || argFiles.get(i).getName().endsWith(".csv"))) {
		        Document doc = builder.build(argFiles.get(i));
		        appendHRs(doc, list);
        	} else {
          	tempe=true;
        	}
        }
        list.sort(null);

        if(tempe) {
          listTemp=new ArrayList<TrkPtInt>();
          File f=argFiles.get(argFiles.size()-1);
          if(f.getName().endsWith(".fit")) decodeTempFIT(f);
          else if(f.getName().endsWith(".csv")) decodeTempCSV(f);
  				listTemp.sort(null);
        }

        Document docTarget = builder.build(argFiles.get(0));
        Element gpx = docTarget.getRootElement();
        Element trk = gpx.getFirstChildElement("trk", "http://www.topografix.com/GPX/1/1");
        Element trkseg = trk.getFirstChildElement("trkseg", "http://www.topografix.com/GPX/1/1");
        Elements trkpts = trkseg.getChildElements("trkpt", "http://www.topografix.com/GPX/1/1");
        
        String svgFilename=argFiles.get(0).getAbsolutePath().replace(".gpx", ".svg");
        PrintWriter svg = new PrintWriter(new File(svgFilename), "UTF-8");
        StringBuilder lines=new StringBuilder();
      	double lon=-1, lat=-1;
      	double lastLon=-1, lastLat=-1;
      	double minLon=Double.MAX_VALUE, minLat=Double.MAX_VALUE, maxLon=Double.MIN_VALUE, maxLat=Double.MIN_VALUE;
      	boolean leaveOut=false;
      	
        TrkPt dummy=new TrkPt(null, null, null);
        TrkPtInt dummyTemp=new TrkPtInt(null, 0);
        for (int i = 0; i < trkpts.size(); i++) {
        	Element trkpt=trkpts.get(i);
        	
        	final double SCALE=10000;
        	lat=Double.parseDouble(trkpt.getAttributeValue("lat"))*SCALE; lon=Double.parseDouble(trkpt.getAttributeValue("lon"))*SCALE;
        	if(i>0) {
	        	lines.append("<line x1='"+lastLon+"' y1='"+-lastLat+"' x2='"+lon+"' y2='"+-lat+"' Xstroke-width='0.0001'");
        	}
        	lastLon=lon; lastLat=lat;
        	minLon=Double.min(minLon, lon); minLat=Double.min(minLat, lat); maxLon=Double.max(maxLon, lon); maxLat=Double.max(maxLat, lat);
        	
          String time = trkpt.getFirstChildElement("time", "http://www.topografix.com/GPX/1/1").getValue();
          Element trackPointExtension=trkpt.getFirstChildElement("extensions", "http://www.topografix.com/GPX/1/1").getFirstChildElement("TrackPointExtension", "http://www.garmin.com/xmlschemas/TrackPointExtension/v1");

          ZonedDateTime zdatetime  = ZonedDateTime.parse(time, DateTimeFormatter.ISO_DATE_TIME);
          dummy.zdt=zdatetime;
          int res=binarySearch(list, dummy);
          TrkPt trkPtFound=null;
          try { // a bit dirty...
          	trkPtFound=list.get(res);
          } catch (IndexOutOfBoundsException e) {}
          if(trkPtFound!=null) {
          	trkPtFound.hr.setNamespacePrefix("gpxtpx");
          	Node hrElem=trkPtFound.hr.copy();
          	trackPointExtension.appendChild(hrElem);
          	
          	if(trkPtFound.temp!=null) {
            	trkPtFound.temp.setNamespacePrefix("gpxtpx");
            	Node tempElem=trkPtFound.temp.copy();
            	trackPointExtension.appendChild(tempElem);
          	}

          	hr = Integer.parseInt(hrElem.getValue());
          	lines.append(" stroke-dasharray='"+(leaveOut?"":"0,")+(double)hr/100.0+(leaveOut?"":",0")+"' Xstroke='hsl("+(120-((hr-hrMin)*120/(hrMax-hrMin)))+", 100%, 50%)' stroke='black'");
          	// TODO https://developer.mozilla.org/en-US/docs/Web/SVG/Attribute/stroke-dashoffset
          	leaveOut=!leaveOut;
          } else {
          	lines.append(" stroke='lightgrey'");
          }
          if(res>middle) {
          	lines.append(" opacity='0.5'");
          }
        	lines.append("/>\n");

          TrkPtInt trkPtIntFound=null;
          if(tempe) {
	        	Instant instant = Instant.parse(time);
	          dummyTemp.zdt=instant;
	          res=binarySearch(listTemp, dummyTemp);
	          try { // a bit dirty...
	          	trkPtIntFound=listTemp.get(res);
	          } catch (IndexOutOfBoundsException e) {}
	          if(trkPtIntFound!=null) {
	          	Element atemp = new Element("gpxtpx:atemp", "http://www.garmin.com/xmlschemas/TrackPointExtension/v1");
	          	atemp.appendChild(trkPtIntFound.value+"");
	            trackPointExtension.appendChild(atemp);
	          }
          }
          
          System.out.println(zdatetime+"\t"+res+"\t"+trkPtFound+"\t"+trkPtIntFound);
        }
        // y flip: https://stackoverflow.com/a/18920936/9717882
      	svg.println("<svg viewBox='"+minLon+" "+-maxLat+" "+(maxLon-minLon)+" "+(maxLat-minLat)+"' xmlns='http://www.w3.org/2000/svg'>");
      	svg.println(lines);
      	svg.println("</svg>");
      	svg.flush(); svg.close();
        //System.out.println(docTarget.toXML());
      	System.out.println(svgFilename);
        
        String docTargetFilename=argFiles.get(0).getAbsolutePath().replace(".gpx", "-"+CombineGPX.class.getSimpleName()+".gpx");
      	PrintWriter pw = new PrintWriter(new File(docTargetFilename), "UTF-8");
        pw.println(docTarget.toXML());
        pw.flush(); pw.close();
        System.out.println(docTargetFilename);
        
        StringSelection selection = new StringSelection(docTargetFilename);
        Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
        clipboard.setContents(selection, selection);
				if (args.length < 2) {
					JOptionPane.showMessageDialog(null, "SVG: "+svgFilename+"\nGPX (copied to clipboard): "+docTargetFilename);
				}
				
        // Keeps entities, but reformats and reindents:
//         FileOutputStream out = new FileOutputStream(args[0]);
//         Serializer ser = new Serializer(out, /*"ISO-8859-1"*/"UTF-8");
//         ser.setIndent(4);
//         ser.write(docTarget);
//         out.flush();
//         out.close();
    }
}
